import os
def init_ssh_keys(inventory):
    import subprocess
    # Run ssh-keygen to generate host keys for login and compute nodes
    key_types = ['rsa','dsa','ecdsa','ed25519']
    #groups = list(filter(lambda x: x != 'hostvars', inventory['all']['children'].keys()))
    groups = ['LoginNodes','ComputeNodes']
    for t in key_types:
        for g in groups:
            output = 'files/ssh_{}_{}_key'.format(g,t)
            if os.path.exists(output):
                continue
            p = subprocess.call(['ssh-keygen','-N','','-f',output,'-t',t])

def init_passwords():
    import yaml
    def gen_password(n=32): # Munge key needs a minimum 32 bytes
        import string
        import random
        characters = string.ascii_letters + string.digits
        return ''.join(random.choice(characters) for i in range(n))
    required_passwords = [
        "mungekey", 
        "slurmdb_passwd", 
        "default_user_password", 
        "influxdb_password", 
    ]
    passwords = {}
    for p in required_passwords:
        passwords[p] = gen_password()
    outputfile = 'vars/passwords.yml'
    if os.path.exists(outputfile):
        return
    with open(outputfile,'w') as f:
        f.write(yaml.dump(passwords))


def make_hosts(inventory):
    # create a file to template /etc/hosts with each nodes IP address
    hostsdata = """
127.0.0.1 localhost
::1     ip6-localhost ip6-loopback
fe00::0 ip6-localnet
ff00::0 ip6-mcastprefix
ff02::1 ip6-allnodes
ff02::2 ip6-allrouters
"""
    for h in inventory['all']['children']['hostvars']['hosts'].items():
        hostsdata = hostsdata + "{} {}\n".format(h[1]['ansible_host'],h[0])
    outputfile = 'files/etcHosts'
    with open(outputfile,'w') as f:
        f.write(hostsdata)

def make_vars_filesystems(choices, inventory):
    import jinja2
    data = {}
    data['domain'] = choices['domain']
    data['project_name'] = choices['project_name']
    project_name = choices['project_name']
    # ATM we have only one NFS server, and its hostname ends in -sql because it does double duty as the SQL server for slurm accounting logs
    nfsserver = "{}-sql0".format(choices['project_name'])

    for vol in ['userdata','slurm_state']:
        data['{}_disk'.format(vol)] = inventory['all']['children']['hostvars']['hosts'][nfsserver]['ansible_host_volumes']["{}_{}".format(project_name,vol)]['dev']

    with open('pre_templates/filesystems_yml.j2') as f:
        template = jinja2.Template(f.read())


    with open('vars/filesystems.yml','w') as f:
        f.write(template.render(**data))

def make_ssh_cfg(choices, inventory):
    import jinja2
    #data = choices | inventory
    data = { **choices, **inventory }
    bastion_server = "{}-bastion0".format(choices['project_name'])
    ansible_user = inventory['all']['children']['hostvars']['hosts'][bastion_server]['ansible_user']
    bastion_floating_ip = inventory['all']['children']['hostvars']['hosts'][bastion_server]['ext_ip']
    data['ansible_user'] = ansible_user
    data['bastion_floating_ip'] = bastion_floating_ip
    with open('pre_templates/ssh_cfg.j2') as f:
        template = jinja2.Template(f.read())
    with open('ssh.cfg','w') as f:
        f.write(template.render(**data))


def make_nfs_exports(inventory):
    # Generate an /etc/exports file for our NFS server to allow compute nodes to connect
    import yaml
    srvopts = "rw,no_root_squash"
    with open('vars/filesystems.yml','r') as f:
        fs = yaml.safe_load(f.read())
    for host in fs['nfsexports']:
        with open('files/etcExports','w') as f:
            for e in host['exportList']:
                f.write('{} {}({})\n'.format(e['src'],'192.168.0.0/24',srvopts))

def groups_from_inventory(inventory):
    groups = {}
    for g in inventory['all']['children'].items():
        groups[g[0]] = g[1]['hosts'].keys()
    return groups

def make_slurm_config(choices, inventory):
    # Generate a slurm.conf from template
    # should include putting the compute nodes in the correct place
    # and listing the mgmt nodes as slurm controlers
    import jinja2
    import yaml
    groups = groups_from_inventory(inventory)
    mgmtnodes = list(inventory['all']['children']['ManagementNodes']['hosts'].keys())
    sqlnodes = list(inventory['all']['children']['SQLNodes']['hosts'].keys())
    mgmtnodes.sort()
    choices['controller']=mgmtnodes[0]
    choices['backup']=mgmtnodes[1]
    choices['sqlnode']=sqlnodes[0]
    with open('pre_templates/slurm_vars.j2') as f:
        template = jinja2.Template(f.read())
    with open('vars/slurm.yml','w') as f:
        f.write(template.render(**choices))

    with open('vars/slurm.yml') as f:
        slurmvars = yaml.safe_load(f.read())
    slurmvars['groups'] = groups
    #slurmvars = slurmvars | choices
    slurmvars = { **slurmvars, **choices }
    with open('pre_templates/slurm.conf.j2') as f:
        template = jinja2.Template(f.read())
    with open('files/slurm.conf','w') as f:
        f.write(template.render(**slurmvars))
    with open('pre_templates/slurmdbd.conf.j2') as f:
        template = jinja2.Template(f.read())
    with open('files/slurmdbd.conf','w') as f:
        f.write(template.render(**slurmvars))

def make_nhc_config(inventory):
    import jinja2
    with open('pre_templates/nhc.conf.j2') as f:
        template = jinja2.Template(f.read())
    with open('files/nhc.conf','w') as f:
        f.write(template.render())

def make_known_hosts(inventory):
    with open('files/ssh_known_hosts','w') as f:
        f.write('\n')
def make_gres_conf(inventory):
    with open('files/gres.conf','w') as f:
        f.write('\n')

def make_vars_vars(choices, inventory):
    import jinja2
    groups = groups_from_inventory(inventory)
    mgmtnodes = list(inventory['all']['children']['ManagementNodes']['hosts'].keys())
    sqlnodes = list(inventory['all']['children']['SQLNodes']['hosts'].keys())
    mgmtnodes.sort()
    choices['controller']=mgmtnodes[0]
    choices['backup']=mgmtnodes[1]
    choices['sqlnode']=sqlnodes[0]
    with open('pre_templates/vars_yml.j2') as f:
        template = jinja2.Template(f.read())
    with open('vars/vars.yml','w') as f:
        f.write(template.render(**choices))
    with open('vars/vars_centos79.yml','w') as f:
        f.write('# empty file for compatibility with HPCasCode\n')



def init_cluster(inventory):
    #
    # These values should not change, but they are chosen at random, so only do it once
    #
    init_ssh_keys(inventory)
    init_passwords()

def derive_ansible_constants(choices, inventory):
    #
    # These values are deviced from the inventory and various choices make about the cluster
    # No matter how many times you run this it won't change
    #
    make_hosts(inventory)
    make_slurm_config(choices, inventory)
    make_vars_filesystems(choices, inventory)
    make_nfs_exports(inventory)
    make_known_hosts(inventory)
    make_nhc_config(inventory)
    make_gres_conf(inventory)
    make_vars_vars(choices, inventory)
    make_ssh_cfg(choices, inventory)


def main():
    import sys
    import yaml
    with open(sys.argv[1]) as f:
        inventory = yaml.safe_load(f.read())
    with open(sys.argv[2]) as f:
        choices = yaml.safe_load(f.read())
    with open(sys.argv[3]) as f:
        versions = yaml.safe_load(f.read())

    choices = { **choices, **versions }

    init_cluster(inventory)

    with open('vars/passwords.yml') as f:
        passwords = yaml.safe_load(f.read())

    choices = { **choices, **passwords }

    derive_ansible_constants(choices, inventory)

if __name__ == "__main__":
    main()
